/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::ArMemExampleClient
 * @author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * @date       2020
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "ArMemExampleClient.h"

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <ArmarXCore/core/time/CycleUtil.h>

#include <RobotAPI/interface/aron.h>

#include <RobotAPI/libraries/armem/component/MemoryRemoteGui.h>
#include <RobotAPI/libraries/armem/client/MemoryWriter.h>
#include <RobotAPI/libraries/armem/client/MemoryReader.h>
#include <RobotAPI/libraries/armem/client/QueryBuilder.h>
#include <RobotAPI/libraries/armem/memory/ice_conversions.h>

#include <RobotAPI/components/armem/ArMemExampleMemory/aron/Primitive.aron.generated.h>
#include <RobotAPI/components/armem/ArMemExampleMemory/aron/SimpleObject.aron.generated.h>


namespace armarx
{
    ArMemExampleClientPropertyDefinitions::ArMemExampleClientPropertyDefinitions(std::string prefix) :
        armarx::ComponentPropertyDefinitions(prefix)
    {
    }

    armarx::PropertyDefinitionsPtr ArMemExampleClient::createPropertyDefinitions()
    {
        armarx::PropertyDefinitionsPtr defs = new ArMemExampleClientPropertyDefinitions(getConfigIdentifier());

        defs->topic(debugObserver);

        defs->optional(memoryName, "mem.MemoryName", "Name of the memory to use.");
        // defs->component(memory, "ArMemExampleMemory");

        return defs;
    }

    std::string ArMemExampleClient::getDefaultName() const
    {
        return "ArMemExampleClient";
    }


    void ArMemExampleClient::onInitComponent()
    {
    }


    void ArMemExampleClient::onConnectComponent()
    {
        createRemoteGuiTab();
        RemoteGui_startRunningTask();

        {
            ARMARX_IMPORTANT << "Waiting for memory '" << memoryName << "'.";
            armem::Time start = armem::Time::now();
            armem::data::WaitForMemoryInput input;
            input.name = memoryName;
            armem::data::WaitForMemoryResult result = memoryNameSystem->waitForMemory(input);
            if (result.success && result.proxy)
            {
                memory = result.proxy;
                usingProxy(memory->ice_getIdentity().name);
            }
            ARMARX_IMPORTANT << "Got memory '" << memoryName << "'"
                             << " (took " << armem::toStringMilliSeconds(armem::Time::now() - start) << ").";
        }
        ARMARX_CHECK_NOT_NULL(memory);

        armem::MemoryID providerID;
        if (true)
        {
            armem::data::AddSegmentInput input;
            input.coreSegmentName = "ExampleModality";
            input.providerSegmentName = "FancyMethodModality";

            ARMARX_IMPORTANT
                    << "Adding segment:"
                    << "\n- core segment:     \t'" << input.coreSegmentName << "'"
                    << "\n- provider segment: \t'" << input.providerSegmentName << "'"
                    ;

            armem::data::AddSegmentsResult results = memory->addSegments({input});
            armem::data::AddSegmentResult& result = results.at(0);

            ARMARX_INFO << "Output: "
                        << "\n- success:      \t" << result.success
                        << "\n- segmentID:    \t" << result.segmentID
                        << "\n- errorMessage: \t" << result.errorMessage
                        ;

            providerID = armem::MemoryID::fromString(result.segmentID);
        }

        armem::MemoryWriter writer(memory);

        armem::MemoryID entityID = providerID;
        entityID.entityName = "example_entity";

        armem::MemoryID snapshotID;
        {
            armem::Commit commit;
            armem::EntityUpdate& update = commit.updates.emplace_back();

            update.entityID = entityID;
            update.timeCreated = armem::Time::now();
            update.instancesData = { new aron::data::AronData(), new aron::data::AronData() };

            ARMARX_IMPORTANT
                    << "Committing:"
                    << "\n- entityID:     \t'" << update.entityID.str() << "'"
                    ;
            armem::CommitResult commitResult = writer.commit(commit);

            ARMARX_CHECK_EQUAL(commitResult.results.size(), 1);
            armem::EntityUpdateResult& result = commitResult.results.at(0);

            ARMARX_INFO << result;

            snapshotID = result.snapshotID;
        }
        {
            armem::Commit commit;
            for (int i = 0; i < 3; ++i)
            {
                armem::EntityUpdate& update = commit.updates.emplace_back();
                update.entityID = entityID;
                update.timeCreated = armem::Time::now() + armem::Time::seconds(i);
                for (int j = 0; j < i; ++j)
                {
                    update.instancesData.push_back(new aron::data::AronData());
                }
            }
            ARMARX_INFO << "Commiting " << commit.updates.size() << " more updates.";

            armem::CommitResult commitResult = writer.commit(commit);
            ARMARX_CHECK_EQUAL(commitResult.results.size(), commit.updates.size());
        }

        if (true)
        {
            queryLatestRawIce(snapshotID.getEntityID());
        }
        if (true)
        {
            querySnapshot(snapshotID);
        }
        if (true)
        {
            primitives(writer);
        }
    }


    void ArMemExampleClient::onDisconnectComponent()
    {
    }


    void ArMemExampleClient::onExitComponent()
    {
    }

    bool ArMemExampleClient::queryLatestRawIce(const armem::MemoryID& entityID)
    {
        ARMARX_IMPORTANT
                << "Querying latest snapshot (via raw ice query types):"
                << "\n- entityID:     \t'" << entityID << "'"
                ;

        armem::data::query::EntityQueryPtr entityQuery =
            new armem::data::query::entity::Single();
        armem::data::query::ProviderSegmentQueryPtr providerQuery =
            new armem::data::query::provider::Single({entityQuery}, entityID.entityName);
        armem::data::query::CoreSegmentQueryPtr coreQuery =
            new armem::data::query::core::Single({providerQuery}, entityID.providerSegmentName);
        armem::data::query::MemoryQueryPtr memoryQuery =
            new armem::data::query::memory::Single({coreQuery}, entityID.coreSegmentName);


        armem::data::QueryInput qInput;
        qInput.memoryQueries = {memoryQuery};

        armem::data::QueryResult qResult = memory->query(qInput);
        ARMARX_INFO << "Result: "
                    << "\n- success:       \t" << qResult.success
                    << "\n- error message: \t" << qResult.errorMessage
                    ;

        if (qResult.success)
        {
            armem::MemoryPtr result = armem::fromIce<armem::MemoryPtr>(qResult.memory);

            ARMARX_CHECK_EQUAL(result->coreSegments.size(), 1);
            armem::CoreSegmentPtr& coreSeg = result->coreSegments.begin()->second;

            ARMARX_CHECK_EQUAL(coreSeg->providerSegments.size(), 1);
            armem::ProviderSegmentPtr& provSeg = coreSeg->providerSegments.begin()->second;

            ARMARX_CHECK_EQUAL(provSeg->entities.size(), 1);
            armem::EntityPtr& entity = provSeg->entities.begin()->second;

            ARMARX_CHECK_GREATER_EQUAL(entity->history.size(), 1);
            armem::EntitySnapshotPtr& snapshot = entity->history.begin()->second;

            ARMARX_INFO << "Result: "
                        << "\n- memory:       \t" << result->name()
                        << "\n- core segment: \t" << coreSeg->name()
                        << "\n- prov segment: \t" << provSeg->name()
                        << "\n- entity:       \t" << entity->name()
                        << "\n- snapshot:     \t" << snapshot->time()
                        << "\n- #instances:   \t" << snapshot->instances.size()
                        ;

            ARMARX_IMPORTANT << "Getting entity from snapshot ID";
            const armem::Entity& e = result->getEntity(entityID);
            ARMARX_INFO << "Result: "
                        << "\n- entity:       \t" << e.name()
                        << "\n- snapshot:     \t" << e.getLatestSnapshot().time()
                        << "\n- #instances:   \t" << e.getLatestSnapshot().instances.size()
                        ;

            tab.queryResult = std::move(result);
            tab.rebuild = true;
        }
        else
        {
            ARMARX_ERROR << qResult.errorMessage;
        }

        return qResult.success;
    }

    bool ArMemExampleClient::querySnapshot(const armem::MemoryID& snapshotID)
    {
        ARMARX_IMPORTANT
                << "Querying exact snapshot of entity: (via query builder)"
                << "\n- snapshotID:     \t'" << snapshotID << "'"
                ;

        armem::MemoryReader reader(memory);

        armem::client::QueryBuilder qb;
        qb
        .coreSegments().withID(snapshotID)
        .providerSegments().withID(snapshotID)
        .entities().withID(snapshotID)
        .snapshots().withID(snapshotID);

        armem::data::QueryResult qResult = reader.query(qb.buildQueryInput());
        ARMARX_INFO << "Result: "
                    << "\n- success:       \t" << qResult.success
                    << "\n- error message: \t" << qResult.errorMessage
                    ;

        if (qResult.success)
        {
            armem::MemoryPtr memory;
            fromIce(qResult.memory, memory);
            armem::EntitySnapshot& entitySnapshot = memory->getEntity(snapshotID).getLatestSnapshot();

            ARMARX_INFO << "Result snapshot: "
                        << "\n- time:        \t" << entitySnapshot.time()
                        << "\n- # instances: \t" << entitySnapshot.instances.size()
                        ;
        }
        else
        {
            ARMARX_ERROR << qResult.errorMessage;
        }

        return qResult.success;
    }

    bool ArMemExampleClient::primitives(armem::MemoryWriter& writer)
    {
        ARMARX_IMPORTANT << "Adding segment " << "Primitive" << "/" << getName();

        auto addSegmentResult = writer.addSegment("Primitive", getName());
        if (!addSegmentResult.success)
        {
            ARMARX_ERROR << addSegmentResult.errorMessage;
            return false;
        }
        armem::MemoryID providerID = armem::MemoryID::fromString(addSegmentResult.segmentID);

        armem::Time time = armem::Time::now();
        armem::Commit commit;

        {
            armem::EntityUpdate& update = commit.updates.emplace_back();
            update.entityID = providerID;
            update.entityID.entityName = "default";
            update.timeCreated = time;

            armem::aron::test::Primitive primitive;
            update.instancesData = { primitive.toAron() };
        }
        {
            armem::EntityUpdate& update = commit.updates.emplace_back();
            update.entityID = providerID;
            update.entityID.entityName = "the answer";
            update.timeCreated = time;

            armem::aron::test::Primitive primitive;
            primitive.the_bool = true;
            primitive.the_double = 2 * 42.42;
            primitive.the_float = 21.5;
            primitive.the_int = 42;
            primitive.the_long = 424242;
            primitive.the_string = "fourty two";
            update.instancesData = { primitive.toAron() };
        }

        armem::CommitResult commitResult = writer.commit(commit);
        if (!commitResult.allSuccess())
        {
            ARMARX_WARNING << commitResult.allErrorMessages();
            return false;
        }

        // Query all entities from provider.
        armem::client::QueryBuilder qb;
        qb
        .coreSegments().withID(providerID)
        .providerSegments().withID(providerID)
        .entities().all()
        .snapshots().all();

        armem::MemoryReader reader(memory);
        armem::data::QueryResult result = reader.query(qb.buildQueryInput());
        if (result.success)
        {
            armem::MemoryPtr resultMemory = armem::fromIce<armem::MemoryPtr>(result.memory);

            tab.queryResult = std::move(resultMemory);
            tab.rebuild = true;
        }
        else
        {
            ARMARX_ERROR << result.errorMessage;
        }

        return true;
    }


    void ArMemExampleClient::createRemoteGuiTab()
    {
        using namespace armarx::RemoteGui::Client;

        if (tab.queryResult)
        {
            armem::MemoryRemoteGui mrg;
            tab.queryResultGroup = mrg.makeGroupBox(*tab.queryResult);
        }

        VBoxLayout root = {tab.queryResultGroup, VSpacer()};
        RemoteGui_createTab(getName(), root, &tab);
    }

    void ArMemExampleClient::RemoteGui_update()
    {
        if (tab.rebuild.exchange(false))
        {
            createRemoteGuiTab();
        }
    }

}
