/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * \package    RobotAPI::gui-plugins::ArMemMemoryViewerWidgetController
 * \author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * \date       2020
 * \copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "ArMemMemoryViewerWidgetController.h"

#include <string>

#include <QSplitter>
#include <QTimer>

#include <ArmarXCore/observers/variant/Variant.h>

#include <RobotAPI/libraries/armem_gui/QTreeWidgetAronDataVisitor.h>


namespace armarx
{
    QString ArMemMemoryViewerWidgetController::GetWidgetName()
    {
        return "MemoryX.ArMemMemoryViewer";
    }


    ArMemMemoryViewerWidgetController::ArMemMemoryViewerWidgetController()
    {
        widget.setupUi(getWidget());

        {
            QSplitter* splitter = new QSplitter();
            while (widget.treesLayout->count() > 0)
            {
                int index = 0; //widget.treesLayout->count() - 1;
                if (widget.treesLayout->itemAt(index)->widget())
                {
                    QLayoutItem* item = widget.treesLayout->takeAt(index);
                    splitter->addWidget(item->widget());
                }
            }
            widget.treesLayout->addWidget(splitter);
        }

        widget.statusLabel->clear();


        // Update timer
        updateTimer = new QTimer(this);
        updateTimer->setInterval(static_cast<int>(std::round(1000 / widget.updateFrequencySpinBox->value())));
        widget.updateFrequencySpinBox->setEnabled(widget.updateAutoCheckBox->isChecked());


        auto replaceWidget = [](auto*& old, QWidget * neu, QLayout * parentLayout)
        {
            QLayoutItem* oldItem = parentLayout->replaceWidget(old, neu);
            if (oldItem)
            {
                delete oldItem;
                delete old;
                old = nullptr;
            }
        };

        memoryTree = new armem::MemoryTreeWidget();
        replaceWidget(widget._memoryTreeWidget, memoryTree, widget.memoryGroupBox->layout());
        ARMARX_CHECK_NULL(widget._memoryTreeWidget);

        queryTabs = new armem::MemoryQueryWidget();
        replaceWidget(widget._queryTabWidget, queryTabs, widget.queryGroupBox->layout());
        ARMARX_CHECK_NULL(widget._queryTabWidget);

        instanceTree = new armem::InstanceTreeWidget();
        instanceTree->setStatusLabel(widget.statusLabel);
        replaceWidget(widget._instanceTree, instanceTree, widget.instanceGroupBox->layout());
        ARMARX_CHECK_NULL(widget._instanceTree);


        connect(this, &This::connected, this, &This::updateMemory);
        connect(widget.updateButton, &QPushButton::pressed, this, &This::updateMemory);
        connect(updateTimer, &QTimer::timeout, this, &This::updateMemory);

        connect(widget.updateAutoCheckBox, &QCheckBox::toggled, widget.updateFrequencySpinBox, &QDoubleSpinBox::setEnabled);
        connect(widget.updateFrequencySpinBox, &QDoubleSpinBox::editingFinished, [this]()
        {
            updateTimer->setInterval(static_cast<int>(std::round(1000 / widget.updateFrequencySpinBox->value())));
        });
        connect(widget.updateAutoCheckBox, &QCheckBox::toggled, this, [this](bool checked)
        {
            if (checked)
            {
                updateTimer->start();
            }
            else
            {
                updateTimer->stop();
            }
        });

        connect(memoryTree, &armem::MemoryTreeWidget::updated, this, &This::memoryTreeUpdated);
        connect(memoryTree, &armem::MemoryTreeWidget::instanceSelected, this, &This::updateInstanceTree);
        connect(instanceTree, &armem::InstanceTreeWidget::updated, this, &This::instanceTreeUpdated);
    }

    ArMemMemoryViewerWidgetController::~ArMemMemoryViewerWidgetController()
    {
    }


    void ArMemMemoryViewerWidgetController::onInitComponent()
    {
        if (memoryName.size() > 0)
        {
            usingProxy(memoryName);
        }
        if (debugObserverName.size() > 0)
        {
            usingProxy(debugObserverName);
        }

        emit inited();
    }

    void ArMemMemoryViewerWidgetController::onConnectComponent()
    {
        if (memoryName.size() > 0)
        {
            getProxy(memory, memoryName);
            memoryReader = armem::MemoryReader(memory);
        }
        // DebugObserver is optional (check for null on every call)
        if (!debugObserverName.empty())
        {
            getProxy(debugObserver, debugObserverName, false, "", false);
        }

        if (widget.updateAutoCheckBox->isChecked())
        {
            updateTimer->stop();
        }

        emit connected();
    }


    void ArMemMemoryViewerWidgetController::onDisconnectComponent()
    {
        if (updateTimer)
        {
            updateTimer->stop();
        }
    }


    void ArMemMemoryViewerWidgetController::updateMemory()
    {
        if (!memoryReader)
        {
            return;
        }

        armem::Time start = armem::Time::now();
        {
            armem::data::query::EntityQueryPtr entityQuery = queryTabs->entityQuery();
            ARMARX_CHECK(entityQuery);
            armem::data::query::ProviderSegmentQueryPtr providerQuery = new armem::data::query::provider::All({entityQuery});
            armem::data::query::CoreSegmentQueryPtr coreQuery = new armem::data::query::core::All({providerQuery});
            armem::data::query::MemoryQueryPtr memoryQuery = new armem::data::query::memory::All({coreQuery});

            this->memoryData = memoryReader.queryToCpp(memoryQuery);
        }
        ARMARX_VERBOSE << "Memory query took " << (armem::Time::now() - start).toMilliSecondsDouble() << " ms.";

        if (debugObserver)
        {
            debugObserver->setDebugDatafield(debugObserverChannelName, "Memory Query [ms]", new Variant((armem::Time::now() - start).toMilliSecondsDouble()));
        }

        if (this->memoryData)
        {
            updateMemoryTree(*this->memoryData);
        }
        else
        {
            widget.statusLabel->setText("No query result.");
        }
    }

    void ArMemMemoryViewerWidgetController::updateInstanceTree(const armem::MemoryID& id)
    {
        if (memoryData)
        {
            instanceTree->update(id, *memoryData);
        }
    }

    void ArMemMemoryViewerWidgetController::updateMemoryTree(armem::Memory& memory)
    {
        armem::Time start = armem::Time::now();

        memoryTree->update(memory);

        ARMARX_VERBOSE << "GUI update took " << (armem::Time::now() - start).toMilliSecondsDouble() << " ms.";
        if (debugObserver)
        {
            debugObserver->setDebugDatafield(debugObserverChannelName, "GUI Update [ms]", new Variant((armem::Time::now() - start).toMilliSecondsDouble()));
        }
    }


    const static std::string CONFIG_KEY_MEMORY = "MemoryName";
    const static std::string CONFIG_KEY_DEBUG_OBSERVER = "DebugObserverName";

    void ArMemMemoryViewerWidgetController::loadSettings(QSettings* settings)
    {
        memoryName = settings->value(QString::fromStdString(CONFIG_KEY_MEMORY),
                                     "Memory").toString().toStdString();
        debugObserverName = settings->value(QString::fromStdString(CONFIG_KEY_DEBUG_OBSERVER),
                                            "DebugObserver").toString().toStdString();
    }
    void ArMemMemoryViewerWidgetController::saveSettings(QSettings* settings)
    {
        settings->setValue(QString::fromStdString(CONFIG_KEY_MEMORY),
                           QString::fromStdString(memoryName));
        settings->setValue(QString::fromStdString(CONFIG_KEY_DEBUG_OBSERVER),
                           QString::fromStdString(debugObserverName));
    }

    QPointer<QDialog> ArMemMemoryViewerWidgetController::getConfigDialog(QWidget* parent)
    {
        if (!configDialog)
        {
            configDialog = new SimpleConfigDialog(parent);
            configDialog->addProxyFinder<armem::MemoryInterfacePrx>({CONFIG_KEY_MEMORY, "Memory", "*Memory"});
            configDialog->addProxyFinder<armarx::DebugObserverInterfacePrx>({CONFIG_KEY_DEBUG_OBSERVER, "Debug Observer", "DebugObserver"});

        }
        return qobject_cast<QDialog*>(configDialog);
    }

    void ArMemMemoryViewerWidgetController::configured()
    {
        if (configDialog)
        {
            memoryName = configDialog->getProxyName(CONFIG_KEY_MEMORY);
            debugObserverName = configDialog->getProxyName(CONFIG_KEY_DEBUG_OBSERVER);
        }
    }
}
