#pragma once

#include "../InternalCommit.h"
#include "../Entity.h"
#include "../EntitySnapshot.h"
#include "../../error/ArMemError.h"

#include "MemoryContainer.h"


namespace armarx::armem::detail
{

    /**
     * @brief A container of entities at some point in the hierarchy.
     *
     * Can be updated by multiple entity updates.
     */
    template <class ValueT, class Derived>
    class EntityContainer : public MemoryContainer<std::map<std::string, ValueT>, Derived>
    {
        using Base = MemoryContainer<std::map<std::string, ValueT>, Derived>;

    public:

        EntityContainer()
        {}
        EntityContainer(const MemoryID& id) :
            MemoryContainer<std::map<std::string, ValueT>, Derived> (id)
        {}

        virtual ~EntityContainer() override
        {}


        virtual const std::string& name() const = 0;
        virtual std::string& name()
        {
            return const_cast<std::string&>(const_cast<const EntityContainer*>(this)->name());
        }


        /**
         * @brief Store all updates in `commit`.
         * @param commit The commit.
         * @return The resulting memory IDs.
         */
        std::vector<MemoryID> update(const InternalCommit& commit)
        {
            std::vector<MemoryID> ids;
            for (const auto& update : commit.updates)
            {
                ids.push_back(this->update(update));
            }
            return ids;
        }
        /**
         * @brief Store the given update.
         * @param update The update.
         * @return The resulting entity snapshot's ID.
         */
        virtual MemoryID update(const InternalEntityUpdate& update) = 0;


        /**
         * @brief Retrieve an entity.
         * @param id The entity ID.
         * @return The entity.
         * @throw An exception deriving from `armem::error::ArMemError` if the entity is missing.
         */
        virtual Entity& getEntity(const MemoryID& id)
        {
            return const_cast<Entity&>(const_cast<const EntityContainer*>(this)->getEntity(id));
        }
        virtual const Entity& getEntity(const MemoryID& id) const = 0;

        /**
         * @brief Retrieve an entity snapshot.
         *
         * Uses `getEntity()` to retrieve the respective entity.
         *
         * @param id The snapshot ID.
         * @return The entity snapshot.
         * @throw An exception deriving from `armem::error::ArMemError` if the snapshot is missing.
         */
        virtual EntitySnapshot& getEntitySnapshot(const MemoryID& id)
        {
            return const_cast<EntitySnapshot&>(const_cast<const EntityContainer*>(this)->getEntitySnapshot(id));
        }
        virtual const EntitySnapshot& getEntitySnapshot(const MemoryID& id) const
        {
            const Entity& entity = getEntity(id);

            if (id.hasTimestamp())
            {
                return entity.getSnapshot(id);
            }
            else
            {
                return entity.getLatestSnapshot();
            }
        }


        virtual EntityInstance& getEntityInstance(const MemoryID& id)
        {
            return getEntitySnapshot(id).getInstance(id);
        }
        virtual const EntityInstance& getEntityInstance(const MemoryID& id) const
        {
            return getEntitySnapshot(id).getInstance(id);
        }


        // MemoryLevel interface
    public:
        /// Return `name`.
        virtual std::string getKeyString() const override
        {
            return name();
        }


    protected:

        /// @throw `armem::error::ContainerNameMismatch` Of `expectedName != actualName`.
        void _checkContainerName(const std::string& expectedName, const std::string& actualName) const
        {
            if (expectedName != actualName)
            {
                throw armem::error::ContainerNameMismatch(expectedName, this->getLevelName(), actualName);
            }
        }

    };

}
