#include "MemoryTreeWidget.h"

#include <QHeaderView>

#include <SimoxUtility/algorithm/string.h>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <RobotAPI/libraries/aron/aroncore/navigators/typenavigator/AronObjectTypeNavigator.h>
#include <RobotAPI/libraries/armem_gui/QTreeWidgetAronDataVisitor.h>


namespace armarx::armem
{

    MemoryTreeWidget::MemoryTreeWidget()
    {
        initBuilders();
        initWidget();
    }

    void MemoryTreeWidget::initWidget()
    {
        clear();
        QStringList columns;
        columns.insert(int(Columns::KEY), "Key");
        columns.insert(int(Columns::SIZE), "#");
        columns.insert(int(Columns::TYPE), "Type");
        columns.insert(int(Columns::LEVEL), "Level");
        columns.insert(int(Columns::ID), "Memory ID");
        setColumnCount(columns.size());
        setHeaderLabels(columns);

        header()->setMinimumSectionSize(25);
        header()->resizeSection(int(Columns::KEY), 250);
        header()->resizeSection(int(Columns::SIZE), 25);
        header()->setTextElideMode(Qt::TextElideMode::ElideRight);


        connect(this, &QTreeWidget::currentItemChanged, this, &MemoryTreeWidget::handleSelection);
    }

    void MemoryTreeWidget::initBuilders()
    {
        memoryBuilder.setExpand(true);
        memoryBuilder.setMakeItemFn([this](const std::string & name, const Memory * memory)
        {
            return makeItem(name, memory);
        });
        memoryBuilder.setUpdateItemFn([this](const std::string&, const Memory * memory, QTreeWidgetItem * memoryItem)
        {
            updateContainerItem(memory, memoryItem);
            if (memoryItem)
            {
                updateChildren(*memory, memoryItem);
            }
            return true;
        });


        coreSegmentBuilder.setExpand(true);
        coreSegmentBuilder.setMakeItemFn([this](const std::string & name, const CoreSegmentPtr & coreSeg)
        {
            return makeItem(name, coreSeg.get());
        });
        coreSegmentBuilder.setUpdateItemFn([this](const std::string&, const CoreSegmentPtr & coreSeg, QTreeWidgetItem * coreSegItem)
        {
            updateTypedContainerItem(coreSeg.get(), coreSegItem);
            if (coreSeg)
            {
                updateChildren(*coreSeg, coreSegItem);
            }
            return true;
        });

        provSegmentBuilder.setExpand(true);
        provSegmentBuilder.setMakeItemFn([this](const std::string & name, const ProviderSegmentPtr & provSeg)
        {
            return makeItem(name, provSeg.get());
        });
        provSegmentBuilder.setUpdateItemFn([this](const std::string&, const ProviderSegmentPtr & provSeg, QTreeWidgetItem * provSegItem)
        {
            updateTypedContainerItem(provSeg.get(), provSegItem);
            if (provSeg)
            {
                updateChildren(*provSeg, provSegItem);
            }
            return true;
        });

        // entityBuilder.setExpand(true);
        entityBuilder.setMakeItemFn([this](const std::string & name, const EntityPtr & entity)
        {
            return makeItem(name, entity.get());
        });
        entityBuilder.setUpdateItemFn([this](const std::string&, const EntityPtr & entity, QTreeWidgetItem * entityItem)
        {
            updateContainerItem(entity.get(), entityItem);
            if (entity)
            {
                updateChildren(*entity, entityItem);
            }
            return true;
        });

        snapshotBuilder.setMakeItemFn([this](const armem::Time & time, const EntitySnapshotPtr & snapshot)
        {
            QTreeWidgetItem* item = makeItem(toDateTimeMilliSeconds(time), snapshot.get());
            item->setData(int(Columns::KEY), Qt::ItemDataRole::UserRole, QVariant(static_cast<qlonglong>(snapshot->time().toMicroSeconds())));
            return item;
        });
        snapshotBuilder.setCompareFn([](const auto & pair, QTreeWidgetItem * item)
        {
            const EntitySnapshotPtr& snapshot = pair.second;
            return armarx::detail::compare(static_cast<qlonglong>(snapshot->time().toMicroSeconds()),
                                           item->data(int(Columns::KEY), Qt::ItemDataRole::UserRole).toLongLong());
        });
        snapshotBuilder.setUpdateItemFn([this](const armem::Time&, const EntitySnapshotPtr & snapshot, QTreeWidgetItem * snapshotItem)
        {
            updateContainerItem(snapshot.get(), snapshotItem);
            if (snapshot)
            {
                updateChildren(*snapshot, snapshotItem);
            }
            else
            {
                while (snapshotItem->childCount() > 0)
                {
                    delete snapshotItem->takeChild(snapshotItem->childCount() - 1);
                }
            }
            return true;
        });

        instanceBuilder.setMakeItemFn([this](const EntityInstancePtr & instance)
        {
            QTreeWidgetItem* item = makeItem("", instance.get());
            item->setData(int(Columns::ID), Qt::UserRole, QString::fromStdString(instance->id().str()));
            return item;
        });
        instanceBuilder.setCompareFn([](const EntityInstancePtr & lhs, QTreeWidgetItem * rhsItem)
        {
            return armarx::detail::compare(lhs->index(), rhsItem->text(0).toInt());
        });
        instanceBuilder.setUpdateItemFn([this](const EntityInstancePtr & instance, QTreeWidgetItem * instanceItem)
        {
            updateItemItem(instance.get(), instanceItem);
            if (instance)
            {
                updateChildren(*instance, instanceItem);
            }
            return true;
        });

    }


    void MemoryTreeWidget::update(const armem::Memory& memory)
    {
        updateChildren(memory, this);
        emit updated();
    }

    void MemoryTreeWidget::update(const std::map<std::string, const armem::Memory*>& memories)
    {
        updateChildren(memories, this);
        emit updated();
    }

    void MemoryTreeWidget::handleSelection()
    {
        QTreeWidgetItem* item = this->currentItem();
        if (!item
            || item->data(int(Columns::LEVEL), Qt::UserRole).toString().toStdString() != EntityInstance().getLevelName())
        {
            return;
        }

        QString itemData = item->data(int(Columns::ID), Qt::UserRole).toString();
        MemoryID instanceID(itemData.toStdString());

        emit instanceSelected(instanceID);
    }


    void MemoryTreeWidget::updateChildren(const armem::Memory& memory, QTreeWidget* tree)
    {
        updateChildren(std::map<std::string, const armem::Memory*> {{memory.name(), &memory}}, tree);
    }

    void MemoryTreeWidget::updateChildren(const std::map<std::string, const armem::Memory*>& memories, QTreeWidget* tree)
    {
        memoryBuilder.updateTree(tree, memories);
    }


    void MemoryTreeWidget::updateChildren(const armem::Memory& memory, QTreeWidgetItem* memoryItem)
    {
        coreSegmentBuilder.updateTree(memoryItem, memory.coreSegments);
    }

    void MemoryTreeWidget::updateChildren(const armem::CoreSegment& coreSeg, QTreeWidgetItem* coreSegItem)
    {
        provSegmentBuilder.updateTree(coreSegItem, coreSeg.providerSegments);
    }

    void MemoryTreeWidget::updateChildren(const armem::ProviderSegment& provSeg, QTreeWidgetItem* provSegItem)
    {
        entityBuilder.updateTree(provSegItem, provSeg.entities);
    }

    void MemoryTreeWidget::updateChildren(const armem::Entity& entity, QTreeWidgetItem* entityItem)
    {
        snapshotBuilder.updateTree(entityItem, entity.history);
    }

    void MemoryTreeWidget::updateChildren(const armem::EntitySnapshot& snapshot, QTreeWidgetItem* snapshotItem)
    {
        instanceBuilder.updateTree(snapshotItem, snapshot.instances);
    }

    void MemoryTreeWidget::updateChildren(const armem::EntityInstance& data, QTreeWidgetItem* dataItem)
    {
        if (instanceIntrospection)
        {
            if (data.dataNavigator())
            {
                QTreeWidgetAronDataVisitor visitor(dataItem);
                visitor.applyTo(*data.dataNavigator());
            }
            else
            {
                dataItem->addChild(new QTreeWidgetItem(QStringList{"(No introspection available.)"}));
            }
        }
    }

    QTreeWidgetItem* MemoryTreeWidget::makeItem(const std::string& key, const armem::detail::MemoryItem* memoryItem)
    {
        QStringList columns;
        if (memoryItem)
        {
            columns.insert(int(Columns::KEY), QString::fromStdString(memoryItem->getKeyString()));
            columns.insert(int(Columns::SIZE), "");
            columns.insert(int(Columns::TYPE), "");
            columns.insert(int(Columns::LEVEL), QString::fromStdString(simox::alg::capitalize_words(memoryItem->getLevelName())));
            columns.insert(int(Columns::ID), QString::fromStdString(memoryItem->id().str()));
        }
        else
        {
            columns.insert(int(Columns::KEY), QString::fromStdString(key));
            columns.insert(int(Columns::ID), QString::fromStdString("(null)"));
        }

        QTreeWidgetItem* item = new QTreeWidgetItem(columns);
        if (memoryItem)
        {
            item->setData(int(Columns::LEVEL), Qt::UserRole, QString::fromStdString(memoryItem->getLevelName()));
        }
        item->setTextAlignment(int(Columns::SIZE), Qt::AlignRight);
        return item;
    }

    void MemoryTreeWidget::updateItemItem(const armem::detail::MemoryItem* level, QTreeWidgetItem* item)
    {
        (void) level, (void) item;
    }

    template <class... T>
    void MemoryTreeWidget::updateContainerItem(
        const armem::detail::MemoryContainer<T...>* container, QTreeWidgetItem* item)
    {
        updateItemItem(container, item);
        if (container)
        {
            item->setText(int(Columns::SIZE), QString::number(container->size()));
        }
    }

    template <class... T>
    void MemoryTreeWidget::updateTypedContainerItem(
        const armem::detail::TypedEntityContainer<T...>* container, QTreeWidgetItem* item)
    {
        updateContainerItem(container, item);
        if (container)
        {
            std::string typeName;
            if (container->aronType())
            {
                auto objectType = aron::type::AronObjectTypePtr::dynamicCast(container->aronType());
                if (objectType)
                {
                    aron::typenavigator::AronObjectTypeNavigator n(objectType, {});
                    typeName = n.getName();
                }
            }
            else
            {
                typeName = "(no Aron type)";
            }
            item->setText(int(Columns::TYPE), QString::fromStdString(typeName));
        }
    }

}
