/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Header
#include "AronOpenCVMatTypeCppSerializer.h"

namespace armarx
{
    namespace aron
    {
        namespace codegeneration
        {
            namespace classwriter
            {
                namespace cppSerializer
                {

                    const std::map<std::string, std::pair<std::string, int>> AronOpenCVMatTypeCppSerializer::ACCEPTED_TYPES =
                    {
                        {"CV_8U", {"CV_8U", 1}},
                        {"CV_8S", {"CV_8S", 1}},
                        {"CV_16U", {"CV_16U", 2}},
                        {"CV_16S", {"CV_16S", 2}},
                        {"CV_32S", {"CV_32S", 4}},
                        {"CV_32F", {"CV_32F", 4}},
                        {"CV_64F", {"CV_64F", 8}},
                    };

                    const std::string AronOpenCVMatTypeCppSerializer::OPENCVMAT_ITERATOR_ACCESSOR = "_iterator";
                    const std::string AronOpenCVMatTypeCppSerializer::OPENCVMAT_DIMENSIONS_ACCESSOR = "_dimensions";
                    const std::string AronOpenCVMatTypeCppSerializer::OPENCVMAT_DIFF_ACCESSOR = "_dimensions";

                    // constructors
                    AronOpenCVMatTypeCppSerializer::AronOpenCVMatTypeCppSerializer(const typenavigator::AronOpenCVMatTypeNavigatorPtr& e) :
                        AronTypeCppSerializer("std::shared_ptr<cv::Mat>", simox::meta::get_type_name(typeid(data::AronNDArray)), simox::meta::get_type_name(typeid(type::AronOpenCVMatType))),
                        typenavigator(e)
                    {
                        // check if type exists
                        ACCEPTED_TYPES.at(e->getUsedType());
                    }

                    std::vector<CppFieldPtr> AronOpenCVMatTypeCppSerializer::getPublicVariableDeclarations(const std::string& name) const
                    {
                        CppFieldPtr field = CppFieldPtr(new CppField(getCppTypename(), name));
                        return {field};
                    }

                    std::vector<std::pair<std::string, std::string>> AronOpenCVMatTypeCppSerializer::getCtorInitializers(const std::string&) const
                    {
                        return {};
                    }

                    CppBlockPtr AronOpenCVMatTypeCppSerializer::getCtorBlock(const std::string&) const
                    {
                        CppBlockPtr b = CppBlockPtr(new CppBlock());
                        return b;
                    }

                    CppBlockPtr AronOpenCVMatTypeCppSerializer::getResetBlock(const std::string& accessor) const
                    {
                        CppBlockPtr b = CppBlockPtr(new CppBlock());

                        std::string escaped_accessor = EscapeAccessor(accessor);
                        std::string accessor_dimensions = escaped_accessor + OPENCVMAT_DIMENSIONS_ACCESSOR;
                        std::string accessor_iterator = escaped_accessor + OPENCVMAT_ITERATOR_ACCESSOR;

                        b->addLine("std::vector<int> " + accessor_dimensions + ";");
                        b->addLine("for (int " + accessor_iterator + " = 0; " + accessor_iterator + " < " + accessor + "->dims; ++" + accessor_iterator + ")");
                        CppBlockPtr b2 = CppBlockPtr(new CppBlock());
                        b2->addLine(accessor_dimensions + ".push_back(" + accessor + "->size[" + accessor_iterator + "]);");
                        b->appendBlock(b2);
                        b->addLine(accessor + " = " + getCppTypename() + "(new cv::Mat(" + accessor_dimensions + ", " + accessor + "->type()));");
                        return b;
                    }

                    CppBlockPtr AronOpenCVMatTypeCppSerializer::getInitializeBlock(const std::string& accessor) const
                    {
                        CppBlockPtr b = CppBlockPtr(new CppBlock());
                        b->addLine(accessor + " = " + getCppTypename() + "(new cv::Mat(std::vector<int>({" + simox::alg::to_string<int>(typenavigator->getDimensions(), ", ") + "}), " + ACCEPTED_TYPES.at(typenavigator->getUsedType()).first + "));");
                        return b;
                    }

                    CppBlockPtr AronOpenCVMatTypeCppSerializer::getWriteInitialTypeBlock(const std::string&) const
                    {
                        CppBlockPtr b = CppBlockPtr(new CppBlock());
                        b->addLine("w.writeOpenCVMatType({" + simox::alg::to_string<int>(typenavigator->getDimensions(), ", ") + "}, \"" + typenavigator->getUsedType() + "\");");
                        return b;
                    }

                    CppBlockPtr AronOpenCVMatTypeCppSerializer::getWriteCurrentTypeBlock(const std::string& accessor) const
                    {
                        CppBlockPtr b = CppBlockPtr(new CppBlock());

                        std::string escaped_accessor = EscapeAccessor(accessor);
                        std::string accessor_dimensions = escaped_accessor + OPENCVMAT_DIMENSIONS_ACCESSOR;
                        std::string accessor_iterator = escaped_accessor + OPENCVMAT_ITERATOR_ACCESSOR;

                        b->addLine("std::vector<int> " + accessor_dimensions + ";");
                        b->addLine("for (int " + accessor_iterator + " = 0; " + accessor_iterator + " < " + accessor + "->dims; ++" + accessor_iterator + ")");
                        CppBlockPtr b2 = CppBlockPtr(new CppBlock());
                        b2->addLine(accessor_dimensions + ".push_back(" + accessor + "->size[" + accessor_iterator + "]);");
                        b->appendBlock(b2);
                        b->addLine("w.writeOpenCVMatType(" + accessor_dimensions + ", std::to_string(" + accessor + "->type()));");
                        return b;
                    }

                    CppBlockPtr AronOpenCVMatTypeCppSerializer::getWriteBlock(const std::string& accessor) const
                    {
                        CppBlockPtr b = CppBlockPtr(new CppBlock());

                        std::string escaped_accessor = EscapeAccessor(accessor);
                        std::string accessor_dimensions = escaped_accessor + OPENCVMAT_DIMENSIONS_ACCESSOR;
                        std::string accessor_iterator = escaped_accessor + OPENCVMAT_ITERATOR_ACCESSOR;

                        b->addLine("std::vector<int> " + accessor_dimensions + ";");
                        b->addLine("for (int " + accessor_iterator + " = 0; " + accessor_iterator + " < " + accessor + "->dims; ++" + accessor_iterator + ")");
                        CppBlockPtr b2 = CppBlockPtr(new CppBlock());
                        b2->addLine(accessor_dimensions + ".push_back(" + accessor + "->size[" + accessor_iterator + "]);");
                        b->appendBlock(b2);
                        b->addLine(accessor_dimensions + ".push_back(" + accessor + "->elemSize());");
                        b->addLine("w.writeNDArray(" + accessor_dimensions + ", std::to_string(" + accessor + "->type()), reinterpret_cast<unsigned char*>(" + accessor + "->data));");
                        return b;
                    }

                    CppBlockPtr AronOpenCVMatTypeCppSerializer::getReadBlock(const std::string& accessor) const
                    {
                        CppBlockPtr b = CppBlockPtr(new CppBlock());
                        b->addLine("const auto [dims, type] = r.readStartNDArray();");
                        b->addLine(accessor + " = " + getCppTypename() + "(new cv::Mat(std::vector<int>({dims.begin(), std::prev(dims.end())}), std::stoi(type)));");
                        b->addLine("r.readEndNDArray(reinterpret_cast<unsigned char*>(" + accessor + "->data));");
                        return b;
                    }

                    CppBlockPtr AronOpenCVMatTypeCppSerializer::getEqualsBlock(const std::string& accessor, const std::string& otherInstanceAccessor) const
                    {
                        CppBlockPtr b = CppBlockPtr(new CppBlock());
                        b->addLine("if (cv::countNonZero(*" + accessor + " != *" + otherInstanceAccessor + ") != 0)");
                        b->addLine("\t return false;");
                        return b;
                    }
                }
            }
        }
    }
}

