/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Header
#include "AronNDArrayDataNavigator.h"

// ArmarX
#include <ArmarXCore/core/exceptions/Exception.h>
#include <RobotAPI/libraries/aron/aroncore/navigators/datanavigator/AronDataNavigatorFactory.h>
#include <RobotAPI/libraries/aron/aroncore/navigators/typenavigator/AronOpenCVMatTypeNavigator.h>
#include <RobotAPI/libraries/aron/aroncore/navigators/typenavigator/AronEigenMatrixTypeNavigator.h>
#include <RobotAPI/libraries/aron/aroncore/navigators/typenavigator/AronIVTCByteImageTypeNavigator.h>
#include <RobotAPI/libraries/aron/aroncore/navigators/typenavigator/AronPCLPointCloudTypeNavigator.h>

namespace armarx
{
    namespace aron
    {
        namespace datanavigator
        {
            // constructors
            AronNDArrayDataNavigator::AronNDArrayDataNavigator(const data::AronNDArrayPtr& o, const AronPath& path) :
                AronNavigator<AronDataDescriptor, data::AronData>::AronNavigator(AronDataDescriptor::eAronNDArray, path),
                AronDataNavigator(AronDataDescriptor::eAronNDArray, path),
                aron(o)
            {
                CheckAronPtrForNull("AronNDArrayDataNavigator", "AronNDArrayDataNavigator", getPath(), aron);
            }

            AronNDArrayDataNavigator::AronNDArrayDataNavigator(const AronPath& path) :
                AronNavigator<AronDataDescriptor, data::AronData>::AronNavigator(AronDataDescriptor::eAronNDArray, path),
                AronDataNavigator(AronDataDescriptor::eAronNDArray, path),
                aron(new data::AronNDArray())
            {

            }

            // static methods
            AronNDArrayDataNavigatorPtr AronNDArrayDataNavigator::DynamicCast(const AronDataNavigatorPtr& n)
            {
                CheckDataNavigatorPtrForNull("AronNDArrayDataNavigator", "DynamicCast[Before]", n);
                AronNDArrayDataNavigatorPtr casted = std::dynamic_pointer_cast<AronNDArrayDataNavigator>(n);
                CheckDataNavigatorPtrForNull("AronNDArrayDataNavigator", "DynamicCast[After]", casted);
                return casted;
            }

            // public member functions
            unsigned char* AronNDArrayDataNavigator::getData() const
            {
                return aron->data.data();
            }

            void AronNDArrayDataNavigator::setData(unsigned int elements, const unsigned char* src)
            {
                aron->data = std::vector<unsigned char>(elements);
                memcpy(aron->data.data(), src, elements);
            }

            std::vector<int> AronNDArrayDataNavigator::getDimensions() const
            {
                return aron->dimensions;
            }

            void AronNDArrayDataNavigator::setDimensions(const std::vector<int>& d)
            {
                aron->dimensions = d;
            }

            void AronNDArrayDataNavigator::addDimension(int i)
            {
                aron->dimensions.push_back(i);
            }

            std::string AronNDArrayDataNavigator::getType() const
            {
                return aron->type;
            }

            void AronNDArrayDataNavigator::setType(const std::string& t)
            {
                if (t.empty())
                {
                    throw exception::AronExceptionWithPathInfo("AronNDArrayDataNavigator", "setType", "The type cannot be empty", getPath());
                }

                aron->type = t;
            }

            // virtual implementations
            data::AronDataPtr AronNDArrayDataNavigator::getResult() const
            {
                return aron;
            }

            std::string AronNDArrayDataNavigator::getName() const
            {
                return "AronNDArray<" + simox::alg::to_string(aron->dimensions, ", ") + ", " + aron->type + ">";
            }

            typenavigator::AronTypeNavigatorPtr AronNDArrayDataNavigator::recalculateType() const
            {
                // We set all to openCVMat because it accepts n-dimensional types
                typenavigator::AronOpenCVMatTypeNavigatorPtr typenav = typenavigator::AronOpenCVMatTypeNavigatorPtr(new typenavigator::AronOpenCVMatTypeNavigator(getPath()));
                typenav->setDimensions(std::vector<int>(aron->dimensions.begin(), std::prev(aron->dimensions.end())));
                typenav->setUsedType(aron->type);
                return typenav;
            }

            bool AronNDArrayDataNavigator::fullfillsType(const typenavigator::AronTypeNavigatorPtr& type) const
            {
                if (!AronResolver::TypesAreCorresponding(type->getDescriptor(), getDescriptor()))
                {
                    return false;
                }

                switch (type->getDescriptor())
                {
#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
case eAron##upperType##Type: \
{ \
    typenavigator::Aron##upperType##TypeNavigatorPtr casted = typenavigator::Aron##upperType##TypeNavigator::DynamicCast(type); \
    if (std::vector<int>(aron->dimensions.begin(), std::prev(aron->dimensions.end())) != casted->getDimensions() || aron->type != casted->getUsedType()) \
    { \
        return false; \
    } \
    break; \
}

                    HANDLE_COMPLEX_TYPES
#undef RUN_ARON_MACRO
                    default:
                        throw exception::AronExceptionWithPathInfo("AronNDArrayDataNavigator", "fullfillsType", "Could not cast a type to an NDArray-Type", getPath());
                }
                return true;
            }

            std::vector<AronDataNavigatorPtr> AronNDArrayDataNavigator::getChildren() const
            {
                return {};
            }

            size_t AronNDArrayDataNavigator::childrenSize() const
            {
                return 0;
            }
        }
    }
}
