/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Header
#include "AronIVTCByteImageTypeNavigator.h"

namespace armarx
{
    namespace aron
    {
        namespace typenavigator
        {
            const std::map<std::string, std::vector<std::string>> AronIVTCByteImageTypeNavigator::ACCEPTED_TYPES =
            {
                {"GrayScale", {"CByteImage::ImageType::eGrayScale"}},
                {"RGB24", {"CByteImage::ImageType::eRGB24"}},
                {"RGB24Split", {"CByteImage::ImageType::eRGB24Split"}}
            };

            // constructors
            AronIVTCByteImageTypeNavigator::AronIVTCByteImageTypeNavigator(const AronPath& path) :
                navigator::AronNavigator<AronTypeDescriptor, type::AronType>::AronNavigator(AronTypeDescriptor::eAronIVTCByteImageType, path),
                AronTypeNavigator(AronTypeDescriptor::eAronIVTCByteImageType, path),
                AronNDArraySerializerTypeNavigator(AronTypeDescriptor::eAronEigenMatrixType, path),
                type(new type::AronIVTCByteImageType({0, 0}, ""))
            {
            }

            AronIVTCByteImageTypeNavigator::AronIVTCByteImageTypeNavigator(const type::AronIVTCByteImageTypePtr& o, const AronPath& path) :
                navigator::AronNavigator<AronTypeDescriptor, type::AronType>::AronNavigator(AronTypeDescriptor::eAronIVTCByteImageType, path),
                AronTypeNavigator(AronTypeDescriptor::eAronIVTCByteImageType, path),
                AronNDArraySerializerTypeNavigator(AronTypeDescriptor::eAronEigenMatrixType, path),
                type(o)
            {
                CheckAronPtrForNull("AronIVTCByteImageTypeNavigator", "AronIVTCByteImageTypeNavigator", getPath(), o);
                if (type->dimensions.size() != 2)
                {
                    throw exception::AronSizeNotValidExceptionException("AronIVTCByteImageTypeNavigator", "AronIVTCByteImageTypeNavigator", "The size of the AronPtr is too small", type->dimensions.size(), 2);
                }
            }

            std::vector<int> AronIVTCByteImageTypeNavigator::getDimensions() const
            {
                return type->dimensions;
            }

            void AronIVTCByteImageTypeNavigator::setDimensions(const std::vector<int>& d)
            {
                if (d.size() != 2)
                {
                    throw exception::AronExceptionWithPathInfo("AronIVTCByteImageTypeNavigator", "setDimensions", "Wrong number of elements in dimension array. The size must be equal to 2. The size was: " + std::to_string(d.size()), getPath());
                }

                type->dimensions = d;
            }

            unsigned int AronIVTCByteImageTypeNavigator::getWidth() const
            {
                return type->dimensions[0];
            }

            unsigned int AronIVTCByteImageTypeNavigator::getHeight() const
            {
                return type->dimensions[1];
            }

            std::string AronIVTCByteImageTypeNavigator::getUsedType() const
            {
                return type->typeName;
            }

            void AronIVTCByteImageTypeNavigator::setWidth(const unsigned int& w)
            {
                if (w == 0)
                {
                    throw exception::AronExceptionWithPathInfo("AronIVTCByteImageTypeNavigator", "setWidth", "The width of an  cannot be 0", getPath());
                }
                type->dimensions[0] = w;
            }

            void AronIVTCByteImageTypeNavigator::setHeight(const unsigned int& h)
            {
                if (h == 0)
                {
                    throw exception::AronExceptionWithPathInfo("AronIVTCByteImageTypeNavigator", "setHeight", "The height of an image cannot be 0", getPath());
                }
                type->dimensions[1] = h;
            }

            void AronIVTCByteImageTypeNavigator::setUsedType(const std::string& u)
            {
                if (u.empty())
                {
                    throw exception::AronExceptionWithPathInfo("AronIVTCByteImageTypeNavigator", "setUsedType", "The useType cannot be empty", getPath());
                }

                for (const auto& [key, vec] : ACCEPTED_TYPES)
                {
                    if (u == key || std::find(vec.begin(), vec.end(), key) != vec.end())
                    {
                        type->typeName = key;
                        return;
                    }
                }
                throw exception::AronStringNotValidExceptionException("AronIVTCByteImageTypeNavigator", "setUsedType", "The type was not found in the ACCEPTED_TYPES", u);
            }

            // static methods
            AronIVTCByteImageTypeNavigatorPtr AronIVTCByteImageTypeNavigator::DynamicCast(const AronTypeNavigatorPtr& n)
            {
                CheckTypeNavigatorPtrForNull("AronIVTCByteImageTypeNavigator", "DynamicCast[Before]", n);
                AronIVTCByteImageTypeNavigatorPtr casted = std::dynamic_pointer_cast<AronIVTCByteImageTypeNavigator>(n);
                CheckTypeNavigatorPtrForNull("AronIVTCByteImageTypeNavigator", "DynamicCast[After]", casted);
                return casted;
            }

            // virtual implementations
            type::AronTypePtr AronIVTCByteImageTypeNavigator::getResult() const
            {
                if (type->dimensions.size() != 2)
                {
                    throw exception::AronExceptionWithPathInfo("AronIVTCByteImageTypeNavigator", "getResult", "The dimension size is wrong. Got size: " + std::to_string(type->dimensions.size()), getPath());
                }
                if (std::any_of(type->dimensions.begin(), type->dimensions.end(), [](int i)
            {
                return i <= 0;
            }))
                {
                    throw exception::AronExceptionWithPathInfo("AronIVTCByteImageTypeNavigator", "getResult", "The dimension size is wrong. At least one empty is <=0", getPath());
                }
                if (type->typeName.empty())
                {
                    throw exception::AronExceptionWithPathInfo("AronIVTCByteImageTypeNavigator", "getResult", "The useType of image is empty.", getPath());
                }
                return type;
            }

            std::string AronIVTCByteImageTypeNavigator::getName() const
            {
                return "AronIVTCByteImageType<" + simox::alg::to_string(type->dimensions, ", ") + ", " + type->typeName + ">";
            }
        }
    }
}

