/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Header
#include "AronOpenCVMatTypeNavigator.h"

namespace armarx
{
    namespace aron
    {
        namespace typenavigator
        {
            const std::map<std::string, std::vector<std::string>> AronOpenCVMatTypeNavigator::ACCEPTED_TYPES =
            {
                {"CV_8U", {"8U"}},
                {"CV_8S", {"8S"}},
                {"CV_16U", {"16U"}},
                {"CV_16S", {"16S"}},
                {"CV_32S", {"32S"}},
                {"CV_32F", {"32F"}},
                {"CV_64F", {"64F"}},
            };

            // constructors
            AronOpenCVMatTypeNavigator::AronOpenCVMatTypeNavigator(const AronPath& path) :
                navigator::AronNavigator<AronTypeDescriptor, type::AronType>::AronNavigator(AronTypeDescriptor::eAronOpenCVMatType, path),
                AronTypeNavigator(AronTypeDescriptor::eAronOpenCVMatType, path),
                AronNDArraySerializerTypeNavigator(AronTypeDescriptor::eAronEigenMatrixType, path),
                type(new type::AronOpenCVMatType({}, ""))
            {
            }

            AronOpenCVMatTypeNavigator::AronOpenCVMatTypeNavigator(const type::AronOpenCVMatTypePtr& o, const AronPath& path) :
                navigator::AronNavigator<AronTypeDescriptor, type::AronType>::AronNavigator(AronTypeDescriptor::eAronOpenCVMatType, path),
                AronTypeNavigator(AronTypeDescriptor::eAronOpenCVMatType, path),
                AronNDArraySerializerTypeNavigator(AronTypeDescriptor::eAronEigenMatrixType, path),
                type(o)
            {
                CheckAronPtrForNull("AronOpenCVMatTypeNavigator", "AronOpenCVMatTypeNavigator", getPath(), o);
            }

            std::string AronOpenCVMatTypeNavigator::getUsedType() const
            {
                return type->typeName;
            }

            std::vector<int> AronOpenCVMatTypeNavigator::getDimensions() const
            {
                return type->dimensions;
            }

            void AronOpenCVMatTypeNavigator::setUsedType(const std::string& u)
            {
                if (u.empty())
                {
                    throw exception::AronExceptionWithPathInfo("AronOpenCVMatTypeNavigator", "setUsedType", "The useType cannot be empty", getPath());
                }

                for (const auto& [key, vec] : ACCEPTED_TYPES)
                {
                    if (u == key || std::find(vec.begin(), vec.end(), key) != vec.end())
                    {
                        type->typeName = key;
                        return;
                    }
                }
                throw exception::AronStringNotValidExceptionException("AronOpenCVMatTypeNavigator", "setUsedType", "The type was not found in the ACCEPTED_TYPES", u);
            }

            void AronOpenCVMatTypeNavigator::setDimensions(const std::vector<int>& d)
            {
                if (d.empty())
                {
                    throw exception::AronExceptionWithPathInfo("AronOpenCVMatTypeNavigator", "setDimensions", "The dimensions of an opencv mat cannot be empty", getPath());
                }
                if (std::any_of(d.begin(), d.end(), [](int i)
            {
                return i < -1;
            }))
                {
                    throw exception::AronExceptionWithPathInfo("AronOpenCVMatTypeNavigator", "setDimensions", "The dimension size is wrong. At least one empty is <-1", getPath());
                }
                type->dimensions = d;
            }


            // static methods
            AronOpenCVMatTypeNavigatorPtr AronOpenCVMatTypeNavigator::DynamicCast(const AronTypeNavigatorPtr& n)
            {
                CheckTypeNavigatorPtrForNull("AronOpenCVMatTypeNavigator", "DynamicCast[Before]", n);
                AronOpenCVMatTypeNavigatorPtr casted = std::dynamic_pointer_cast<AronOpenCVMatTypeNavigator>(n);
                CheckTypeNavigatorPtrForNull("AronOpenCVMatTypeNavigator", "DynamicCast[After]", casted);
                return casted;
            }

            // virtual implementations
            type::AronTypePtr AronOpenCVMatTypeNavigator::getResult() const
            {
                if (type->dimensions.empty())
                {
                    throw exception::AronExceptionWithPathInfo("AronOpenCVMatTypeNavigator", "getResult", "The dimension size is empty", getPath());
                }
                if (std::any_of(type->dimensions.begin(), type->dimensions.end(), [](int i)
            {
                return i < -1;
            }))
                {
                    throw exception::AronExceptionWithPathInfo("AronOpenCVMatTypeNavigator", "getResult", "The dimension size is wrong. At least one empty is <-1", getPath());
                }
                if (type->typeName.empty())
                {
                    throw exception::AronExceptionWithPathInfo("AronOpenCVMatTypeNavigator", "getResult", "The useType of image is empty.", getPath());
                }
                return type;
            }

            std::string AronOpenCVMatTypeNavigator::getName() const
            {
                return "AronOpenCVMatType<" + simox::alg::to_string(type->dimensions, ", ") + ", " + type->typeName + ">";
            }
        }
    }
}

